//-----------------------------------------------------------------------------
// Wireless M-Bus communication
//-----------------------------------------------------------------------------
#IFDEF NOT MBUS_INC THEN
#DEFINE MBUS_INC

#DEFINE MBUS_MODULE_MEDIA  1
#DEFINE MBUS_MODULE_PATH   "B:\SYSTEM\EXT\MOD_MBUS.RMX"

#DEFINE MBUS_DATA_LENGTH   246

#DEFINE MBUS_MODCALL_ERR   -999

#DEFINE MBUS_MODE_S2       0
#DEFINE MBUS_MODE_T1       1
#DEFINE MBUS_MODE_T2       2
#DEFINE MBUS_MODE_S1       3
#DEFINE MBUS_MODE_R        4
#DEFINE MBUS_MODE_T1_C     10
#DEFINE MBUS_MODE_T2_C     11


//----------------------------------------------------------------------------
// mbusFrame
//
// Description:
//    This struct block contains a frame received from the M-Bus module.
//
// Output:
//    control      - The control field.
//    manufacturer - The ID of the manufacturer. Part of the device address.
//    id           - The device id/serial number.  Part of the device address.
//    version      - The device version. Part of the device address.
//    type         - The device type. Part of the device address.
//    RSSI         - The signal strength for the packet. If it is not enabled in mbusOpen, it is set to 0.
//    linsec       - RTCU time when frame was received
//    length       - The length of the additional data in the frame.
//    data         - The additional data that was received. 
//                   Only the first <length> bytes are valid.
//
//----------------------------------------------------------------------------
STRUCT_BLOCK ALIGN mbusFrame;
   control      : SINT;           // Control field
   
   // Address
   manufacturer : INT;           // Manufacture ID
   id           : DINT;          // Device Identification number
   version      : SINT;          // Device Version number
   type         : SINT;          // Device Type

   // APP_DATA
   RSSI         : INT;            // signal strength in dB = - RSSI/2
   linsec       : DINT;           // RTCU time when frame was received
   length       : INT;            // Length of additional data
   data         : ARRAY[1 .. MBUS_DATA_LENGTH] OF SINT;  
END_STRUCT_BLOCK;

//----------------------------------------------------------------------------
// mbusInit()
//
// Description:
//    Initializes the M-Bus extension module.
//
// Input:
//    path       - The path to the module to initialize. Optional, as the default value is already correct.
//    media      - The media the module is located on. Optional, as the default value is already correct.
//
// Returns:
//        0   : OK
//       -1   : Failed to open media
//       -2   : Failed to load module
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusInit : INT;
VAR_INPUT
   path     : STRING := MBUS_MODULE_PATH;
   media    : SINT := MBUS_MODULE_MEDIA;
END_VAR;
VAR
   rc       : INT;
END_VAR;

// Initialize media where module is located
rc := fsMediaOpen(media := media);
IF NOT (rc = 0) THEN
   mbusInit := -1;
   RETURN;
END_IF;

// Load an extension module
rc := extModuleLoad(path:=path);
IF NOT (rc = 1) THEN
   mbusInit := -2;
   RETURN;
END_IF;

mbusInit := 0;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusOpen()
//
// Description:
//    Opens the connection to the M-Bus module.
//
// Input:
//    mode       - The M-Bus mode to use. See the MBUS_MODE_* defines above.
//    RSSI       - Include the signal strength( RSSI) for each valid packet.
//                 It can be read from mbusReceive()
//    max_frames - Amount of elements the buffer will accept. Maximum value 1000.
//    overwrite  - Policy to apply when the buffer is full.
//                   If TRUE, remove oldest element in buffer and append new.
//                   If FALSE, discard newly received frame.
//
//
// Returns:
//        0: Success
//       -1: Failed to open the M-Bus interface.
//       -2: Invalid input
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusOpen : INT;
VAR_INPUT
   mode        : SINT := MBUS_MODE_T1;
   RSSI        : BOOL := OFF;
   max_frames  : INT  := 100;
   overwrite   : BOOL := TRUE;
END_VAR;
VAR
   error       : INT;
END_VAR;
   mbusOpen := INT(MODCALL("mod_mbus", "mbusOpen", error));
   IF NOT (error = 0) THEN
      mbusOpen := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusClose()
//
// Description:
//    Closes the connection to the M-Bus module.
//
// Input:
//    None.
//
// Returns:
//        0: Success
//       -1: Failed to close the M-Bus interface.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusClose : INT;
VAR
   error    : INT;
END_VAR;
   mbusClose := INT(MODCALL("mod_mbus", "mbusClose", error));
   IF NOT (error = 0) THEN
      mbusClose := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusSend()
//
// Description:
//    Sends data to the M-Bus module.
//
// Input:
//    control    - The M-Bus mode to use. See the MBUS_MODE_* defines above. 0-255.
//    length     - The length of the data to send. Max 246 bytes.
//    data       - The data to send.
//
// Returns:
//        0: Success
//       -1: Invalid parameter.
//       -2: Failed to send data.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusSend : INT;
VAR_INPUT
   control     : INT := -1;   // (-1, 0 .. 255) :change control byte before send
   length      : INT := 0;    // (0 .. MBUS_DATA_LENGTH) : length of data to send
   data        : PTR;         // data to send
END_VAR;
VAR
   error       : INT;
END_VAR;
   mbusSend := INT(MODCALL("mod_mbus", "mbusSend", error));
   IF NOT (error = 0) THEN
      mbusSend := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusFilterReceive()
//
// Description:
//    Receives data from a registered filter in the M-Bus module.
//
// Input:
//    idx        - (1..64) filter index to receive from.
//
// Output:
//    frame      - The provided struct is filled with the values of a received M-Bus frame. 
//
// Returns:
//        1: Data was received.
//        0: No data
//       -1: Failed to access receive interface.
//       -2: Interface is not open
//       -3: Invalid input
//       -4: No slave registered with this index
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusFilterReceive;
VAR_INPUT
   idx   : INT := 0;
   frame : ACCESS mbusFrame;
END_VAR;
VAR
   error       : INT;
END_VAR;
   mbusFilterReceive := INT(MODCALL("mod_mbus", "mbusFilterReceive", error));
   IF NOT (error = 0) THEN
      mbusFilterReceive := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusReceive()
//
// Description:
//    Receives data from the M-Bus module. If the device address is given, it searches for 
//       data received from a slave device with the given address parameters.
//       If type = -1 (default), this function returns the oldest element in the buffer.
//
// Input:
//    manufacturer - The ID of the manufacturer. Part of the device address.
//    id           - The device identification number/serial number.  Part of the device address.
//    version      - The device version. Part of the device address.
//    type         - The device type. Part of the device address.
//    timeout      - Time to wait for data in milliseconds. 
//                      If timeout is -1 (default), wait until data is received.
//
// Output:
//    frame        - The provided struct is filled with the values of a received M-Bus frame. 
//
// Returns:
//        1: Data was received.
//        0: No data
//       -1: Failed to access receive interface.
//       -2: Interface is not open
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusReceive;
VAR_INPUT
   timeout       : DINT    := -1;
   manufacturer  : INT     := 0;
   id            : DINT    := 0;
   version       : SINT    := 0;
   type          : SINT    := -1;
   frame         : ACCESS mbusFrame;
END_VAR;
VAR
   error          : INT;
END_VAR;
   mbusReceive := INT(MODCALL("mod_mbus", "mbusReceive", error));
   IF NOT (error = 0) THEN
      mbusReceive := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;   

//----------------------------------------------------------------------------
// mbusInfo()
//
// Description:
//    Sends some information about the M-Bus module to the device output.
// Example output:
//
//    mbus_test:Signal quality = 0
//    mbus_test:RSSI           = 205 (-102 dBm)
//    mbus_test:Temperature    = 32 C
//    mbus_test:VCC            = 3450mV
//    mbus_test:MBUS_MODE      = 0x01
//    mbus_test:DATA_INTERFACE = 0x00
//    mbus_test:PART_NUMBER    = RC1180-MBUS3
//    mbus_test:HW_REV_NO      = 2.00
//    mbus_test:FW_REV_NO      = 3.15
//    mbus_info:SERIAL_NUMBER  = 01 23 45 67 89 AB CD EF
//    mbus_info:INSTALL_MODE   = 0x02
//    mbus_info:RSSI_MODE      = 0x01
//
// Input:
//    None.
//
// Returns:
//        0: Success
//       -1: Failed to access interface.
//       -2: Module failure
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION mbusInfo : INT;
VAR
   error    : INT;
END_VAR;
   mbusInfo := INT(MODCALL("mod_mbus", "mbusInfo", error));
   IF NOT (error = 0) THEN
      mbusInfo := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusRegisterSlave()
//
// Description:
//    Registers/installs a slave device on the M-Bus module.
//    This is necessary when filtering or when using encryption.
//
// Input:
//    idx          - The index of the slave to register. 1-64.
//    manufacturer - The ID of the manufacturer. Part of the device address.
//    id           - The device identification number/serial number.  Part of the device address.
//    version      - The device version. Part of the device address.
//    type         - The device type. Part of the device address.
//    key          - The 16 byte (128 bit) encryption key. Can be left empty if encryption is not used.
//
// Returns:
//        0: Success
//       -1: Failed to access interface.
//       -2: Invalid idx.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION mbusRegisterSlave : INT;
VAR_INPUT
   idx           : INT := 1;
   manufacturer  : MANDATORY INT;
   id            : MANDATORY DINT;
   version       : MANDATORY SINT;
   type          : MANDATORY SINT;
   key           : PTR;
END_VAR;
VAR
   error    : INT;
END_VAR;
   mbusRegisterSlave := INT(MODCALL("mod_mbus", "mbusRegisterSlave", error));
   IF NOT (error = 0) THEN
      mbusRegisterSlave := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusFilterEnable()
//
// Description:
//    Enables/disables filtering of the received packets, to only receive packets from registered slave devices.
//
// Input:
//    enable    - Set to true to only receive packets from installed slaves.
//
// Returns:
//        0: Success
//       -1: Failed to access interface.
//       -2: Invalid idx.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION mbusFilterEnable : INT;
VAR_INPUT
   enable           : BOOL := TRUE;
END_VAR;
VAR
   error    : INT;
END_VAR;
   mbusFilterEnable := INT(MODCALL("mod_mbus", "mbusFilterEnable", error));
   IF NOT (error = 0) THEN
      mbusFilterEnable := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusGetBufferLevel()
//
// Description:
//    Returns how many promille of the buffer is filled.
//
// Input:
//
// Returns: Buffer full level in promille [0..1000].
//       -1: Failed to access interface.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusGetBufferLevel : INT;
VAR
   error    : INT;
END_VAR;
   mbusGetBufferLevel := INT(MODCALL("mod_mbus", "mbusGetBufferLevel", error));
   IF NOT (error = 0) THEN
      mbusGetBufferLevel := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

#END_IF